@@ -16,27 +16,26 @@
 
 from collections import deque
 
-from bson.py3compat import integer_types
+from bson import _convert_raw_document_lists_to_streams
+from pymongo.cursor import _SocketManager, _CURSOR_CLOSED_ERRORS
 from pymongo.errors import (ConnectionFailure,
                             InvalidOperation,
-                            NotMasterError,
                             OperationFailure)
 from pymongo.message import (_CursorAddress,
                              _GetMore,
                              _RawBatchGetMore)
+from pymongo.response import PinnedResponse
 
 
 class CommandCursor(object):
     """A cursor / iterator over command cursors."""
     _getmore_class = _GetMore
 
-    def __init__(self, collection, cursor_info, address, retrieved=0,
+    def __init__(self, collection, cursor_info, address,
                  batch_size=0, max_await_time_ms=None, session=None,
                  explicit_session=False):
-        """Create a new command cursor.
-
-        The parameter 'retrieved' is unused.
-        """
+        """Create a new command cursor."""
+        self.__sock_mgr = None
         self.__collection = collection
         self.__id = cursor_info['id']
         self.__data = deque(cursor_info['firstBatch'])
@@ -57,13 +56,12 @@ class CommandCursor(object):
 
         self.batch_size(batch_size)
 
-        if (not isinstance(max_await_time_ms, integer_types)
+        if (not isinstance(max_await_time_ms, int)
                 and max_await_time_ms is not None):
             raise TypeError("max_await_time_ms must be an integer or None")
 
     def __del__(self):
-        if self.__id and not self.__killed:
-            self.__die()
+        self.__die()
 
     def __die(self, synchronous=False):
         """Closes this cursor.
@@ -71,16 +69,23 @@ class CommandCursor(object):
         already_killed = self.__killed
         self.__killed = True
         if self.__id and not already_killed:
+            cursor_id = self.__id
             address = _CursorAddress(
-                self.__address, self.__collection.full_name)
-            if synchronous:
-                self.__collection.database.client._close_cursor_now(
-                    self.__id, address, session=self.__session)
-            else:
-                # The cursor will be closed later in a different session.
-                self.__collection.database.client._close_cursor(
-                    self.__id, address)
-        self.__end_session(synchronous)
+                self.__address, self.__ns)
+        else:
+            # Skip killCursors.
+            cursor_id = 0
+            address = None
+        self.__collection.database.client._cleanup_cursor(
+            synchronous,
+            cursor_id,
+            address,
+            self.__sock_mgr,
+            self.__session,
+            self.__explicit_session)
+        if not self.__explicit_session:
+            self.__session = None
+        self.__sock_mgr = None
 
     def __end_session(self, synchronous):
         if self.__session and not self.__explicit_session:
@@ -108,7 +113,7 @@ class CommandCursor(object):
         :Parameters:
           - `batch_size`: The size of each batch of results requested.
         """
-        if not isinstance(batch_size, integer_types):
+        if not isinstance(batch_size, int):
             raise TypeError("batch_size must be an integer")
         if batch_size < 0:
             raise ValueError("batch_size must be >= 0")
@@ -127,52 +132,59 @@ class CommandCursor(object):
         changeStream aggregate or getMore."""
         return self.__postbatchresumetoken
 
+    def _maybe_pin_connection(self, sock_info):
+        client = self.__collection.database.client
+        if not client._should_pin_cursor(self.__session):
+            return
+        if not self.__sock_mgr:
+            sock_info.pin_cursor()
+            sock_mgr = _SocketManager(sock_info, False)
+            # Ensure the connection gets returned when the entire result is
+            # returned in the first batch.
+            if self.__id == 0:
+                sock_mgr.close()
+            else:
+                self.__sock_mgr = sock_mgr
+
     def __send_message(self, operation):
         """Send a getmore message and handle the response.
         """
-        def kill():
-            self.__killed = True
-            self.__end_session(True)
-
         client = self.__collection.database.client
         try:
-            response = client._run_operation_with_response(
+            response = client._run_operation(
                 operation, self._unpack_response, address=self.__address)
-        except OperationFailure:
-            kill()
-            raise
-        except NotMasterError:
-            # Don't send kill cursors to another server after a "not master"
-            # error. It's completely pointless.
-            kill()
+        except OperationFailure as exc:
+            if exc.code in _CURSOR_CLOSED_ERRORS:
+                # Don't send killCursors because the cursor is already closed.
+                self.__killed = True
+            # Return the session and pinned connection, if necessary.
+            self.close()
             raise
         except ConnectionFailure:
-            # Don't try to send kill cursors on another socket
-            # or to another server. It can cause a _pinValue
-            # assertion on some server releases if we get here
-            # due to a socket timeout.
-            kill()
+            # Don't send killCursors because the cursor is already closed.
+            self.__killed = True
+            # Return the session and pinned connection, if necessary.
+            self.close()
             raise
         except Exception:
-            # Close the cursor
-            self.__die()
+            self.close()
             raise
 
-        from_command = response.from_command
-        reply = response.data
-        docs = response.docs
-
-        if from_command:
-            cursor = docs[0]['cursor']
+        if isinstance(response, PinnedResponse):
+            if not self.__sock_mgr:
+                self.__sock_mgr = _SocketManager(response.socket_info,
+                                                 response.more_to_come)
+        if response.from_command:
+            cursor = response.docs[0]['cursor']
             documents = cursor['nextBatch']
             self.__postbatchresumetoken = cursor.get('postBatchResumeToken')
             self.__id = cursor['id']
         else:
-            documents = docs
-            self.__id = reply.cursor_id
+            documents = response.docs
+            self.__id = response.data.cursor_id
 
         if self.__id == 0:
-            kill()
+            self.close()
         self.__data = deque(documents)
 
     def _unpack_response(self, response, cursor_id, codec_options,
@@ -203,10 +215,9 @@ class CommandCursor(object):
                                     self.__session,
                                     self.__collection.database.client,
                                     self.__max_await_time_ms,
-                                    False))
+                                    self.__sock_mgr, False))
         else:  # Cursor id is zero nothing else to return
-            self.__killed = True
-            self.__end_session(True)
+            self.__die(True)
 
         return len(self.__data)
 
@@ -269,8 +280,7 @@ class CommandCursor(object):
         if not len(self.__data) and not self.__killed and get_more_allowed:
             self._refresh()
         if len(self.__data):
-            coll = self.__collection
-            return coll.database._fix_outgoing(self.__data.popleft(), coll)
+            return self.__data.popleft()
         else:
             return None
 
@@ -284,7 +294,7 @@ class CommandCursor(object):
 class RawBatchCommandCursor(CommandCursor):
     _getmore_class = _RawBatchGetMore
 
-    def __init__(self, collection, cursor_info, address, retrieved=0,
+    def __init__(self, collection, cursor_info, address,
                  batch_size=0, max_await_time_ms=None, session=None,
                  explicit_session=False):
         """Create a new cursor / iterator over raw batches of BSON data.
@@ -293,16 +303,22 @@ class RawBatchCommandCursor(CommandCursor):
         see :meth:`~pymongo.collection.Collection.aggregate_raw_batches`
         instead.
 
-        .. mongodoc:: cursors
+        .. seealso:: The MongoDB documentation on `cursors <https://dochub.mongodb.org/core/cursors>`_.
         """
         assert not cursor_info.get('firstBatch')
         super(RawBatchCommandCursor, self).__init__(
-            collection, cursor_info, address, retrieved, batch_size,
+            collection, cursor_info, address, batch_size,
             max_await_time_ms, session, explicit_session)
 
     def _unpack_response(self, response, cursor_id, codec_options,
                          user_fields=None, legacy_response=False):
-        return response.raw_response(cursor_id)
+        raw_response = response.raw_response(
+            cursor_id, user_fields=user_fields)
+        if not legacy_response:
+            # OP_MSG returns firstBatch/nextBatch documents as a BSON array
+            # Re-assemble the array of documents into a document stream
+            _convert_raw_document_lists_to_streams(raw_response[0])
+        return raw_response
 
     def __getitem__(self, index):
         raise InvalidOperation("Cannot call __getitem__ on RawBatchCursor")
