@@ -1,4 +1,4 @@
-# Copyright 2011-2015 MongoDB, Inc.
+# Copyright 2011-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you
 # may not use this file except in compliance with the License.  You
@@ -15,22 +15,28 @@
 
 """Functions and classes common to multiple pymongo modules."""
 
-import collections
 import datetime
 import warnings
 
-from bson.binary import (STANDARD, PYTHON_LEGACY,
-                         JAVA_LEGACY, CSHARP_LEGACY)
-from bson.codec_options import CodecOptions
-from bson.py3compat import string_type, integer_types, iteritems
+from collections import abc, OrderedDict
+from urllib.parse import unquote_plus
+
+from bson import SON
+from bson.binary import UuidRepresentation
+from bson.codec_options import CodecOptions, TypeRegistry
 from bson.raw_bson import RawBSONDocument
 from pymongo.auth import MECHANISMS
+from pymongo.compression_support import (validate_compressors,
+                                         validate_zlib_compression_level)
+from pymongo.driver_info import DriverInfo
+from pymongo.server_api import ServerApi
 from pymongo.errors import ConfigurationError
 from pymongo.monitoring import _validate_event_listeners
 from pymongo.read_concern import ReadConcern
 from pymongo.read_preferences import _MONGOS_MODES, _ServerMode
-from pymongo.ssl_support import validate_cert_reqs
-from pymongo.write_concern import WriteConcern
+from pymongo.write_concern import DEFAULT_WRITE_CONCERN, WriteConcern
+
+ORDERED_TYPES = (SON, OrderedDict)
 
 # Defaults until we connect to a server and get updated limits.
 MAX_BSON_SIZE = 16 * (1024 ** 2)
@@ -40,13 +46,15 @@ MAX_WIRE_VERSION = 0
 MAX_WRITE_BATCH_SIZE = 1000
 
 # What this version of PyMongo supports.
-MIN_SUPPORTED_WIRE_VERSION = 0
-MAX_SUPPORTED_WIRE_VERSION = 5
+MIN_SUPPORTED_SERVER_VERSION = "3.6"
+MIN_SUPPORTED_WIRE_VERSION = 6
+MAX_SUPPORTED_WIRE_VERSION = 14
 
-# Frequency to call ismaster on servers, in seconds.
+# Frequency to call hello on servers, in seconds.
 HEARTBEAT_FREQUENCY = 10
 
-# Frequency to process kill-cursors, in seconds. See MongoClient.close_cursor.
+# Frequency to clean up unclosed cursors, in seconds.
+# See MongoClient._process_kill_cursors.
 KILL_CURSOR_FREQUENCY = 1
 
 # Frequency to process events queue, in seconds.
@@ -58,9 +66,12 @@ EVENTS_QUEUE_FREQUENCY = 1
 # longest it is willing to wait for a new primary to be found.
 SERVER_SELECTION_TIMEOUT = 30
 
-# Spec requires at least 500ms between ismaster calls.
+# Spec requires at least 500ms between hello calls.
 MIN_HEARTBEAT_INTERVAL = 0.5
 
+# Spec requires at least 60s between SRV rescans.
+MIN_SRV_RESCAN_INTERVAL = 60
+
 # Default connectTimeout in seconds.
 CONNECT_TIMEOUT = 20.0
 
@@ -70,23 +81,40 @@ MAX_POOL_SIZE = 100
 # Default value for minPoolSize.
 MIN_POOL_SIZE = 0
 
+# The maximum number of concurrent connection creation attempts per pool.
+MAX_CONNECTING = 2
+
 # Default value for maxIdleTimeMS.
 MAX_IDLE_TIME_MS = None
 
+# Default value for maxIdleTimeMS in seconds.
+MAX_IDLE_TIME_SEC = None
+
+# Default value for waitQueueTimeoutMS in seconds.
+WAIT_QUEUE_TIMEOUT = None
+
 # Default value for localThresholdMS.
 LOCAL_THRESHOLD_MS = 15
 
-# mongod/s 2.6 and above return code 59 when a
-# command doesn't exist. mongod versions previous
-# to 2.6 and mongos 2.4.x return no error code
-# when a command does exist. mongos versions previous
-# to 2.4.0 return code 13390 when a command does not
-# exist.
-COMMAND_NOT_FOUND_CODES = (59, 13390, None)
+# Default value for retryWrites.
+RETRY_WRITES = True
+
+# Default value for retryReads.
+RETRY_READS = True
+
+# mongod/s 2.6 and above return code 59 when a command doesn't exist.
+COMMAND_NOT_FOUND_CODES = (59,)
 
 # Error codes to ignore if GridFS calls createIndex on a secondary
 UNAUTHORIZED_CODES = (13, 16547, 16548)
 
+# Maximum number of sessions to send in a single endSessions command.
+# From the driver sessions spec.
+_MAX_END_SESSIONS = 10000
+
+# Default value for srvServiceName
+SRV_SERVICE_NAME = "mongodb"
+
 
 def partition_node(node):
     """Split a host:port string into (host, int(port)) pair."""
@@ -101,13 +129,13 @@ def partition_node(node):
 
 
 def clean_node(node):
-    """Split and normalize a node name from an ismaster response."""
+    """Split and normalize a node name from a hello response."""
     host, port = partition_node(node)
 
     # Normalize hostname to lowercase, since DNS is case-insensitive:
     # http://tools.ietf.org/html/rfc4343
     # This prevents useless rediscovery if "foo.com" is in the seed list but
-    # "FOO.com" is in the ismaster response.
+    # "FOO.com" is in the hello response.
     return host.lower(), port
 
 
@@ -118,10 +146,11 @@ def raise_config_error(key, dummy):
 
 # Mapping of URI uuid representation options to valid subtypes.
 _UUID_REPRESENTATIONS = {
-    'standard': STANDARD,
-    'pythonLegacy': PYTHON_LEGACY,
-    'javaLegacy': JAVA_LEGACY,
-    'csharpLegacy': CSHARP_LEGACY
+    'unspecified': UuidRepresentation.UNSPECIFIED,
+    'standard': UuidRepresentation.STANDARD,
+    'pythonLegacy': UuidRepresentation.PYTHON_LEGACY,
+    'javaLegacy': UuidRepresentation.JAVA_LEGACY,
+    'csharpLegacy': UuidRepresentation.CSHARP_LEGACY
 }
 
 
@@ -134,7 +163,7 @@ def validate_boolean(option, value):
 
 def validate_boolean_or_string(option, value):
     """Validates that value is True, False, 'true', or 'false'."""
-    if isinstance(value, string_type):
+    if isinstance(value, str):
         if value not in ('true', 'false'):
             raise ValueError("The value of %s must be "
                              "'true' or 'false'" % (option,))
@@ -145,13 +174,14 @@ def validate_boolean_or_string(option, value):
 def validate_integer(option, value):
     """Validates that 'value' is an integer (or basestring representation).
     """
-    if isinstance(value, integer_types):
+    if isinstance(value, int):
         return value
-    elif isinstance(value, string_type):
-        if not value.isdigit():
+    elif isinstance(value, str):
+        try:
+            return int(value)
+        except ValueError:
             raise ValueError("The value of %s must be "
                              "an integer" % (option,))
-        return int(value)
     raise TypeError("Wrong type for %s, value must be an integer" % (option,))
 
 
@@ -204,13 +234,12 @@ def validate_non_negative_integer_or_none(option, value):
 
 
 def validate_string(option, value):
-    """Validates that 'value' is an instance of `basestring` for Python 2
-    or `str` for Python 3.
+    """Validates that 'value' is an instance of `str`.
     """
-    if isinstance(value, string_type):
+    if isinstance(value, str):
         return value
-    raise TypeError("Wrong type for %s, value must be "
-                    "an instance of %s" % (option, string_type.__name__))
+    raise TypeError("Wrong type for %s, value must be an instance of "
+                    "str" % (option,))
 
 
 def validate_string_or_none(option, value):
@@ -224,16 +253,32 @@ def validate_string_or_none(option, value):
 def validate_int_or_basestring(option, value):
     """Validates that 'value' is an integer or string.
     """
-    if isinstance(value, integer_types):
+    if isinstance(value, int):
         return value
-    elif isinstance(value, string_type):
-        if value.isdigit():
+    elif isinstance(value, str):
+        try:
             return int(value)
-        return value
+        except ValueError:
+            return value
     raise TypeError("Wrong type for %s, value must be an "
                     "integer or a string" % (option,))
 
 
+def validate_non_negative_int_or_basestring(option, value):
+    """Validates that 'value' is an integer or string.
+    """
+    if isinstance(value, int):
+        return value
+    elif isinstance(value, str):
+        try:
+            val = int(value)
+        except ValueError:
+            return value
+        return validate_non_negative_integer(option, val)
+    raise TypeError("Wrong type for %s, value must be an "
+                    "non negative integer or a string" % (option,))
+
+
 def validate_positive_float(option, value):
     """Validates that 'value' is a float, or can be converted to one, and is
        positive.
@@ -285,6 +330,16 @@ def validate_timeout_or_zero(option, value):
     return validate_positive_float(option, value) / 1000.0
 
 
+def validate_timeout_or_none_or_zero(option, value):
+    """Validates a timeout specified in milliseconds returning
+    a value in floating point seconds. value=0 and value="0" are treated the
+    same as value=None which means unlimited timeout.
+    """
+    if value is None or value == 0 or value == "0":
+        return None
+    return validate_positive_float(option, value) / 1000.0
+
+
 def validate_max_staleness(option, value):
     """Validates maxStalenessSeconds according to the Max Staleness Spec."""
     if value == -1 or value == "-1":
@@ -302,7 +357,7 @@ def validate_read_preference(dummy, value):
 
 
 def validate_read_preference_mode(dummy, value):
-    """Validate read preference mode for a MongoReplicaSetClient.
+    """Validate read preference mode for a MongoClient.
 
     .. versionchanged:: 3.5
        Returns the original ``value`` instead of the validated read preference
@@ -316,10 +371,7 @@ def validate_read_preference_mode(dummy, value):
 def validate_auth_mechanism(option, value):
     """Validate the authMechanism URI option.
     """
-    # CRAM-MD5 is for server testing only. Undocumented,
-    # unsupported, may be removed at any time. You have
-    # been warned.
-    if value not in MECHANISMS and value != 'CRAM-MD5':
+    if value not in MECHANISMS:
         raise ValueError("%s must be in %s" % (option, tuple(MECHANISMS)))
     return value
 
@@ -347,8 +399,11 @@ def validate_read_preference_tags(name, value):
             tag_sets.append({})
             continue
         try:
-            tag_sets.append(dict([tag.split(":")
-                                  for tag in tag_set.split(",")]))
+            tags = {}
+            for tag in tag_set.split(","):
+                key, val = tag.split(":")
+                tags[unquote_plus(key)] = unquote_plus(val)
+            tag_sets.append(tags)
         except Exception:
             raise ValueError("%r not a valid "
                              "value for %s" % (tag_set, name))
@@ -357,7 +412,8 @@ def validate_read_preference_tags(name, value):
 
 _MECHANISM_PROPS = frozenset(['SERVICE_NAME',
                               'CANONICALIZE_HOST_NAME',
-                              'SERVICE_REALM'])
+                              'SERVICE_REALM',
+                              'AWS_SESSION_TOKEN'])
 
 
 def validate_auth_mechanism_properties(option, value):
@@ -368,6 +424,10 @@ def validate_auth_mechanism_properties(option, value):
         try:
             key, val = opt.split(':')
         except ValueError:
+            # Try not to leak the token.
+            if 'AWS_SESSION_TOKEN' in opt:
+                opt = ('AWS_SESSION_TOKEN:<redacted token>, did you forget '
+                       'to percent-escape the token with quote_plus?')
             raise ValueError("auth mechanism properties must be "
                              "key:value pairs like SERVICE_NAME:"
                              "mongodb, not %s." % (opt,))
@@ -378,31 +438,61 @@ def validate_auth_mechanism_properties(option, value):
         if key == 'CANONICALIZE_HOST_NAME':
             props[key] = validate_boolean_or_string(key, val)
         else:
-            props[key] = val
+            props[key] = unquote_plus(val)
 
     return props
 
 
 def validate_document_class(option, value):
     """Validate the document_class option."""
-    if not issubclass(value, (collections.MutableMapping, RawBSONDocument)):
+    if not issubclass(value, (abc.MutableMapping, RawBSONDocument)):
         raise TypeError("%s must be dict, bson.son.SON, "
                         "bson.raw_bson.RawBSONDocument, or a "
                         "sublass of collections.MutableMapping" % (option,))
     return value
 
 
+def validate_type_registry(option, value):
+    """Validate the type_registry option."""
+    if value is not None and not isinstance(value, TypeRegistry):
+        raise TypeError("%s must be an instance of %s" % (
+            option, TypeRegistry))
+    return value
+
+
+def validate_list(option, value):
+    """Validates that 'value' is a list."""
+    if not isinstance(value, list):
+        raise TypeError("%s must be a list" % (option,))
+    return value
+
+
+def validate_list_or_none(option, value):
+    """Validates that 'value' is a list or None."""
+    if value is None:
+        return value
+    return validate_list(option, value)
+
+
+def validate_list_or_mapping(option, value):
+    """Validates that 'value' is a list or a document."""
+    if not isinstance(value, (abc.Mapping, list)):
+        raise TypeError("%s must either be a list or an instance of dict, "
+                        "bson.son.SON, or any other type that inherits from "
+                        "collections.Mapping" % (option,))
+
+
 def validate_is_mapping(option, value):
     """Validate the type of method arguments that expect a document."""
-    if not isinstance(value, collections.Mapping):
+    if not isinstance(value, abc.Mapping):
         raise TypeError("%s must be an instance of dict, bson.son.SON, or "
-                        "other type that inherits from "
+                        "any other type that inherits from "
                         "collections.Mapping" % (option,))
 
 
 def validate_is_document_type(option, value):
     """Validate the type of method arguments that expect a MongoDB document."""
-    if not isinstance(value, (collections.MutableMapping, RawBSONDocument)):
+    if not isinstance(value, (abc.MutableMapping, RawBSONDocument)):
         raise TypeError("%s must be an instance of dict, bson.son.SON, "
                         "bson.raw_bson.RawBSONDocument, or "
                         "a type that inherits from "
@@ -420,6 +510,33 @@ def validate_appname_or_none(option, value):
     return value
 
 
+def validate_driver_or_none(option, value):
+    """Validate the driver keyword arg."""
+    if value is None:
+        return value
+    if not isinstance(value, DriverInfo):
+        raise TypeError("%s must be an instance of DriverInfo" % (option,))
+    return value
+
+
+def validate_server_api_or_none(option, value):
+    """Validate the server_api keyword arg."""
+    if value is None:
+        return value
+    if not isinstance(value, ServerApi):
+        raise TypeError("%s must be an instance of ServerApi" % (option,))
+    return value
+
+
+def validate_is_callable_or_none(option, value):
+    """Validates that 'value' is a callable."""
+    if value is None:
+        return value
+    if not callable(value):
+        raise ValueError("%s must be a callable" % (option,))
+    return value
+
+
 def validate_ok_for_replace(replacement):
     """Validate a replacement document."""
     validate_is_mapping("replacement", replacement)
@@ -432,12 +549,14 @@ def validate_ok_for_replace(replacement):
 
 def validate_ok_for_update(update):
     """Validate an update document."""
-    validate_is_mapping("update", update)
-    # Update can not be {}
+    validate_list_or_mapping("update", update)
+    # Update cannot be {}.
     if not update:
-        raise ValueError('update only works with $ operators')
+        raise ValueError('update cannot be empty')
+
+    is_document = not isinstance(update, list)
     first = next(iter(update))
-    if not first.startswith('$'):
+    if is_document and not first.startswith('$'):
         raise ValueError('update only works with $ operators')
 
 
@@ -462,65 +581,143 @@ def validate_tzinfo(dummy, value):
     return value
 
 
-# journal is an alias for j,
-# wtimeoutms is an alias for wtimeout,
-URI_VALIDATORS = {
-    'replicaset': validate_string_or_none,
-    'w': validate_int_or_basestring,
-    'wtimeout': validate_integer,
-    'wtimeoutms': validate_integer,
-    'fsync': validate_boolean_or_string,
-    'j': validate_boolean_or_string,
+def validate_auto_encryption_opts_or_none(option, value):
+    """Validate the driver keyword arg."""
+    if value is None:
+        return value
+    from pymongo.encryption_options import AutoEncryptionOpts
+    if not isinstance(value, AutoEncryptionOpts):
+        raise TypeError("%s must be an instance of AutoEncryptionOpts" % (
+            option,))
+
+    return value
+
+
+# Dictionary where keys are the names of public URI options, and values
+# are lists of aliases for that option.
+URI_OPTIONS_ALIAS_MAP = {
+    'tls': ['ssl'],
+}
+
+# Dictionary where keys are the names of URI options, and values
+# are functions that validate user-input values for that option. If an option
+# alias uses a different validator than its public counterpart, it should be
+# included here as a key, value pair.
+URI_OPTIONS_VALIDATOR_MAP = {
+    'appname': validate_appname_or_none,
+    'authmechanism': validate_auth_mechanism,
+    'authmechanismproperties': validate_auth_mechanism_properties,
+    'authsource': validate_string,
+    'compressors': validate_compressors,
+    'connecttimeoutms': validate_timeout_or_none_or_zero,
+    'directconnection': validate_boolean_or_string,
+    'heartbeatfrequencyms': validate_timeout_or_none,
     'journal': validate_boolean_or_string,
-    'maxpoolsize': validate_positive_integer_or_none,
-    'socketkeepalive': validate_boolean_or_string,
-    'waitqueuemultiple': validate_non_negative_integer_or_none,
-    'ssl': validate_boolean_or_string,
-    'ssl_keyfile': validate_readable,
-    'ssl_certfile': validate_readable,
-    'ssl_pem_passphrase': validate_string_or_none,
-    'ssl_cert_reqs': validate_cert_reqs,
-    'ssl_ca_certs': validate_readable,
-    'ssl_match_hostname': validate_boolean_or_string,
-    'ssl_crlfile': validate_readable,
+    'localthresholdms': validate_positive_float_or_zero,
+    'maxidletimems': validate_timeout_or_none,
+    'maxconnecting': validate_positive_integer,
+    'maxpoolsize': validate_non_negative_integer_or_none,
+    'maxstalenessseconds': validate_max_staleness,
     'readconcernlevel': validate_string_or_none,
     'readpreference': validate_read_preference_mode,
     'readpreferencetags': validate_read_preference_tags,
-    'localthresholdms': validate_positive_float_or_zero,
-    'authmechanism': validate_auth_mechanism,
-    'authsource': validate_string,
-    'authmechanismproperties': validate_auth_mechanism_properties,
-    'tz_aware': validate_boolean_or_string,
-    'uuidrepresentation': validate_uuid_representation,
-    'connect': validate_boolean_or_string,
-    'minpoolsize': validate_non_negative_integer,
-    'appname': validate_appname_or_none,
-    'unicode_decode_error_handler': validate_unicode_decode_error_handler
-}
-
-TIMEOUT_VALIDATORS = {
-    'connecttimeoutms': validate_timeout_or_none,
-    'sockettimeoutms': validate_timeout_or_none,
-    'waitqueuetimeoutms': validate_timeout_or_none,
+    'replicaset': validate_string_or_none,
+    'retryreads': validate_boolean_or_string,
+    'retrywrites': validate_boolean_or_string,
+    'loadbalanced': validate_boolean_or_string,
     'serverselectiontimeoutms': validate_timeout_or_zero,
-    'heartbeatfrequencyms': validate_timeout_or_none,
-    'maxidletimems': validate_timeout_or_none,
-    'maxstalenessseconds': validate_max_staleness,
+    'sockettimeoutms': validate_timeout_or_none_or_zero,
+    'tls': validate_boolean_or_string,
+    'tlsallowinvalidcertificates': validate_boolean_or_string,
+    'tlsallowinvalidhostnames': validate_boolean_or_string,
+    'tlscafile': validate_readable,
+    'tlscertificatekeyfile': validate_readable,
+    'tlscertificatekeyfilepassword': validate_string_or_none,
+    'tlsdisableocspendpointcheck': validate_boolean_or_string,
+    'tlsinsecure': validate_boolean_or_string,
+    'w': validate_non_negative_int_or_basestring,
+    'wtimeoutms': validate_non_negative_integer,
+    'zlibcompressionlevel': validate_zlib_compression_level,
+    'srvservicename': validate_string,
+    'srvmaxhosts': validate_non_negative_integer
 }
 
+# Dictionary where keys are the names of URI options specific to pymongo,
+# and values are functions that validate user-input values for those options.
+NONSPEC_OPTIONS_VALIDATOR_MAP = {
+    'connect': validate_boolean_or_string,
+    'driver': validate_driver_or_none,
+    'server_api': validate_server_api_or_none,
+    'fsync': validate_boolean_or_string,
+    'minpoolsize': validate_non_negative_integer,
+    'tlscrlfile': validate_readable,
+    'tz_aware': validate_boolean_or_string,
+    'unicode_decode_error_handler': validate_unicode_decode_error_handler,
+    'uuidrepresentation': validate_uuid_representation,
+    'waitqueuemultiple': validate_non_negative_integer_or_none,
+    'waitqueuetimeoutms': validate_timeout_or_none,
+}
+
+# Dictionary where keys are the names of keyword-only options for the
+# MongoClient constructor, and values are functions that validate user-input
+# values for those options.
 KW_VALIDATORS = {
     'document_class': validate_document_class,
+    'type_registry': validate_type_registry,
     'read_preference': validate_read_preference,
     'event_listeners': _validate_event_listeners,
     'tzinfo': validate_tzinfo,
     'username': validate_string_or_none,
     'password': validate_string_or_none,
+    'server_selector': validate_is_callable_or_none,
+    'auto_encryption_opts': validate_auto_encryption_opts_or_none,
 }
 
-URI_VALIDATORS.update(TIMEOUT_VALIDATORS)
-VALIDATORS = URI_VALIDATORS.copy()
+# Dictionary where keys are any URI option name, and values are the
+# internally-used names of that URI option. Options with only one name
+# variant need not be included here. Options whose public and internal
+# names are the same need not be included here.
+INTERNAL_URI_OPTION_NAME_MAP = {
+    'ssl': 'tls',
+}
+
+# Map from deprecated URI option names to a tuple indicating the method of
+# their deprecation and any additional information that may be needed to
+# construct the warning message.
+URI_OPTIONS_DEPRECATION_MAP = {
+    # format: <deprecated option name>: (<mode>, <message>),
+    # Supported <mode> values:
+    # - 'renamed': <message> should be the new option name. Note that case is
+    #   preserved for renamed options as they are part of user warnings.
+    # - 'removed': <message> may suggest the rationale for deprecating the
+    #   option and/or recommend remedial action.
+    # For example:
+    # 'wtimeout': ('renamed', 'wTimeoutMS'),
+}
+
+# Augment the option validator map with pymongo-specific option information.
+URI_OPTIONS_VALIDATOR_MAP.update(NONSPEC_OPTIONS_VALIDATOR_MAP)
+for optname, aliases in URI_OPTIONS_ALIAS_MAP.items():
+    for alias in aliases:
+        if alias not in URI_OPTIONS_VALIDATOR_MAP:
+            URI_OPTIONS_VALIDATOR_MAP[alias] = (
+                URI_OPTIONS_VALIDATOR_MAP[optname])
+
+# Map containing all URI option and keyword argument validators.
+VALIDATORS = URI_OPTIONS_VALIDATOR_MAP.copy()
 VALIDATORS.update(KW_VALIDATORS)
 
+# List of timeout-related options.
+TIMEOUT_OPTIONS = [
+    'connecttimeoutms',
+    'heartbeatfrequencyms',
+    'maxidletimems',
+    'maxstalenessseconds',
+    'serverselectiontimeoutms',
+    'sockettimeoutms',
+    'waitqueuetimeoutms',
+]
+
 
 _AUTH_OPTIONS = frozenset(['authmechanismproperties'])
 
@@ -532,7 +729,7 @@ def validate_auth_option(option, value):
     if lower not in _AUTH_OPTIONS:
         raise ConfigurationError('Unknown '
                                  'authentication option: %s' % (option,))
-    return lower, value
+    return option, value
 
 
 def validate(option, value):
@@ -541,29 +738,45 @@ def validate(option, value):
     lower = option.lower()
     validator = VALIDATORS.get(lower, raise_config_error)
     value = validator(option, value)
-    return lower, value
+    return option, value
 
 
 def get_validated_options(options, warn=True):
     """Validate each entry in options and raise a warning if it is not valid.
-    Returns a copy of options with invalid entries removed
+    Returns a copy of options with invalid entries removed.
+
+    :Parameters:
+        - `opts`: A dict containing MongoDB URI options.
+        - `warn` (optional): If ``True`` then warnings will be logged and
+          invalid options will be ignored. Otherwise, invalid options will
+          cause errors.
     """
-    validated_options = {}
-    for opt, value in iteritems(options):
-        lower = opt.lower()
+    if isinstance(options, _CaseInsensitiveDictionary):
+        validated_options = _CaseInsensitiveDictionary()
+        get_normed_key = lambda x: x
+        get_setter_key = lambda x: options.cased_key(x)
+    else:
+        validated_options = {}
+        get_normed_key = lambda x: x.lower()
+        get_setter_key = lambda x: x
+
+    for opt, value in options.items():
+        normed_key = get_normed_key(opt)
         try:
-            validator = URI_VALIDATORS.get(lower, raise_config_error)
+            validator = URI_OPTIONS_VALIDATOR_MAP.get(
+                normed_key, raise_config_error)
             value = validator(opt, value)
-        except (ValueError, ConfigurationError) as exc:
+        except (ValueError, TypeError, ConfigurationError) as exc:
             if warn:
                 warnings.warn(str(exc))
             else:
                 raise
         else:
-            validated_options[lower] = value
+            validated_options[get_setter_key(normed_key)] = value
     return validated_options
 
 
+# List of write-concern-related options.
 WRITE_CONCERN_OPTIONS = frozenset([
     'w',
     'wtimeout',
@@ -622,6 +835,14 @@ class BaseObject(object):
         """
         return self.__write_concern
 
+    def _write_concern_for(self, session):
+        """Read only access to the write concern of this instance or session.
+        """
+        # Override this operation's write concern with the transaction's.
+        if session and session.in_transaction:
+            return DEFAULT_WRITE_CONCERN
+        return self.write_concern
+
     @property
     def read_preference(self):
         """Read only access to the read preference of this instance.
@@ -631,10 +852,99 @@ class BaseObject(object):
         """
         return self.__read_preference
 
+    def _read_preference_for(self, session):
+        """Read only access to the read preference of this instance or session.
+        """
+        # Override this operation's read preference with the transaction's.
+        if session:
+            return session._txn_read_preference() or self.__read_preference
+        return self.__read_preference
+
     @property
     def read_concern(self):
-        """Read only access to the read concern of this instance.
+        """Read only access to the :class:`~pymongo.read_concern.ReadConcern`
+        of this instance.
 
         .. versionadded:: 3.2
         """
         return self.__read_concern
+
+
+class _CaseInsensitiveDictionary(abc.MutableMapping):
+    def __init__(self, *args, **kwargs):
+        self.__casedkeys = {}
+        self.__data = {}
+        self.update(dict(*args, **kwargs))
+
+    def __contains__(self, key):
+        return key.lower() in self.__data
+
+    def __len__(self):
+        return len(self.__data)
+
+    def __iter__(self):
+        return (key for key in self.__casedkeys)
+
+    def __repr__(self):
+        return str({self.__casedkeys[k]: self.__data[k] for k in self})
+
+    def __setitem__(self, key, value):
+        lc_key = key.lower()
+        self.__casedkeys[lc_key] = key
+        self.__data[lc_key] = value
+
+    def __getitem__(self, key):
+        return self.__data[key.lower()]
+
+    def __delitem__(self, key):
+        lc_key = key.lower()
+        del self.__casedkeys[lc_key]
+        del self.__data[lc_key]
+
+    def __eq__(self, other):
+        if not isinstance(other, abc.Mapping):
+            return NotImplemented
+        if len(self) != len(other):
+            return False
+        for key in other:
+            if self[key] != other[key]:
+                return False
+
+        return True
+
+    def get(self, key, default=None):
+        return self.__data.get(key.lower(), default)
+
+    def pop(self, key, *args, **kwargs):
+        lc_key = key.lower()
+        self.__casedkeys.pop(lc_key, None)
+        return self.__data.pop(lc_key, *args, **kwargs)
+
+    def popitem(self):
+        lc_key, cased_key = self.__casedkeys.popitem()
+        value = self.__data.pop(lc_key)
+        return cased_key, value
+
+    def clear(self):
+        self.__casedkeys.clear()
+        self.__data.clear()
+
+    def setdefault(self, key, default=None):
+        lc_key = key.lower()
+        if key in self:
+            return self.__data[lc_key]
+        else:
+            self.__casedkeys[lc_key] = key
+            self.__data[lc_key] = default
+            return default
+
+    def update(self, other):
+        if isinstance(other, _CaseInsensitiveDictionary):
+            for key in other:
+                self[other.cased_key(key)] = other[key]
+        else:
+            for key in other:
+                self[key] = other[key]
+
+    def cased_key(self, key):
+        return self.__casedkeys[key.lower()]
