@@ -17,47 +17,49 @@
 import sys
 import traceback
 
-from bson.py3compat import abc, iteritems, itervalues, string_type
+from collections import abc
+
 from bson.son import SON
 from pymongo import ASCENDING
 from pymongo.errors import (CursorNotFound,
                             DuplicateKeyError,
                             ExecutionTimeout,
-                            NotMasterError,
+                            NotPrimaryError,
                             OperationFailure,
                             WriteError,
                             WriteConcernError,
                             WTimeoutError)
+from pymongo.hello import HelloCompat
 
 # From the SDAM spec, the "node is shutting down" codes.
 _SHUTDOWN_CODES = frozenset([
     11600,  # InterruptedAtShutdown
     91,     # ShutdownInProgress
 ])
-# From the SDAM spec, the "not master" error codes are combined with the
+# From the SDAM spec, the "not primary" error codes are combined with the
 # "node is recovering" error codes (of which the "node is shutting down"
 # errors are a subset).
-_NOT_MASTER_CODES = frozenset([
-    10107,  # NotMaster
-    13435,  # NotMasterNoSlaveOk
+_NOT_PRIMARY_CODES = frozenset([
+    10058,  # LegacyNotPrimary <=3.2 "not primary" error code
+    10107,  # NotWritablePrimary
+    13435,  # NotPrimaryNoSecondaryOk
     11602,  # InterruptedDueToReplStateChange
-    13436,  # NotMasterOrSecondary
+    13436,  # NotPrimaryOrSecondary
     189,    # PrimarySteppedDown
 ]) | _SHUTDOWN_CODES
 # From the retryable writes spec.
-_RETRYABLE_ERROR_CODES = _NOT_MASTER_CODES | frozenset([
+_RETRYABLE_ERROR_CODES = _NOT_PRIMARY_CODES | frozenset([
     7,     # HostNotFound
     6,     # HostUnreachable
     89,    # NetworkTimeout
     9001,  # SocketException
     262,   # ExceededTimeLimit
 ])
-_UUNDER = u"_"
 
 
 def _gen_index_name(keys):
     """Generate an index name from the set of fields it is over."""
-    return _UUNDER.join(["%s_%s" % item for item in keys])
+    return "_".join(["%s_%s" % item for item in keys])
 
 
 def _index_list(key_or_list, direction=None):
@@ -68,8 +70,10 @@ def _index_list(key_or_list, direction=None):
     if direction is not None:
         return [(key_or_list, direction)]
     else:
-        if isinstance(key_or_list, string_type):
+        if isinstance(key_or_list, str):
             return [(key_or_list, ASCENDING)]
+        if isinstance(key_or_list, abc.ItemsView):
+            return list(key_or_list)
         elif not isinstance(key_or_list, (list, tuple)):
             raise TypeError("if no direction is specified, "
                             "key_or_list must be an instance of list")
@@ -84,7 +88,7 @@ def _index_document(index_list):
     if isinstance(index_list, abc.Mapping):
         raise TypeError("passing a dict to sort/create_index/hint is not "
                         "allowed - use a list of tuples instead. did you "
-                        "mean %r?" % list(iteritems(index_list)))
+                        "mean %r?" % list(index_list.items()))
     elif not isinstance(index_list, (list, tuple)):
         raise TypeError("must use a list of (key, direction) pairs, "
                         "not: " + repr(index_list))
@@ -93,16 +97,17 @@ def _index_document(index_list):
 
     index = SON()
     for (key, value) in index_list:
-        if not isinstance(key, string_type):
-            raise TypeError("first item in each key pair must be a string")
-        if not isinstance(value, (string_type, int, abc.Mapping)):
+        if not isinstance(key, str):
+            raise TypeError(
+                "first item in each key pair must be an instance of str")
+        if not isinstance(value, (str, int, abc.Mapping)):
             raise TypeError("second item in each key pair must be 1, -1, "
                             "'2d', or another valid MongoDB index specifier.")
         index[key] = value
     return index
 
 
-def _check_command_response(response, max_wire_version, msg=None,
+def _check_command_response(response, max_wire_version,
                             allowable_errors=None,
                             parse_write_concern_error=False):
     """Check the response to a command for errors.
@@ -115,92 +120,54 @@ def _check_command_response(response, max_wire_version, msg=None,
                                max_wire_version)
 
     if parse_write_concern_error and 'writeConcernError' in response:
-        _raise_write_concern_error(response['writeConcernError'])
+        _error = response["writeConcernError"]
+        _labels = response.get("errorLabels")
+        if _labels:
+            _error.update({'errorLabels': _labels})
+        _raise_write_concern_error(_error)
 
-    if not response["ok"]:
+    if response["ok"]:
+        return
 
-        details = response
-        # Mongos returns the error details in a 'raw' object
-        # for some errors.
-        if "raw" in response:
-            for shard in itervalues(response["raw"]):
-                # Grab the first non-empty raw error from a shard.
-                if shard.get("errmsg") and not shard.get("ok"):
-                    details = shard
-                    break
-
-        errmsg = details["errmsg"]
-        if (allowable_errors is None
-                or (errmsg not in allowable_errors
-                    and details.get("code") not in allowable_errors)):
-
-            code = details.get("code")
-            # Server is "not master" or "recovering"
-            if code in _NOT_MASTER_CODES:
-                raise NotMasterError(errmsg, response)
-            elif ("not master" in errmsg
-                  or "node is recovering" in errmsg):
-                raise NotMasterError(errmsg, response)
-
-            # Server assertion failures
-            if errmsg == "db assertion failure":
-                errmsg = ("db assertion failure, assertion: '%s'" %
-                          details.get("assertion", ""))
-                raise OperationFailure(errmsg,
-                                       details.get("assertionCode"),
-                                       response,
-                                       max_wire_version)
-
-            # Other errors
-            # findAndModify with upsert can raise duplicate key error
-            if code in (11000, 11001, 12582):
-                raise DuplicateKeyError(errmsg, code, response,
-                                        max_wire_version)
-            elif code == 50:
-                raise ExecutionTimeout(errmsg, code, response,
-                                       max_wire_version)
-            elif code == 43:
-                raise CursorNotFound(errmsg, code, response,
-                                     max_wire_version)
-
-            msg = msg or "%s"
-            raise OperationFailure(msg % errmsg, code, response,
-                                   max_wire_version)
-
-
-def _check_gle_response(result, max_wire_version):
-    """Return getlasterror response as a dict, or raise OperationFailure."""
-    # Did getlasterror itself fail?
-    _check_command_response(result, max_wire_version)
-
-    if result.get("wtimeout", False):
-        # MongoDB versions before 1.8.0 return the error message in an "errmsg"
-        # field. If "errmsg" exists "err" will also exist set to None, so we
-        # have to check for "errmsg" first.
-        raise WTimeoutError(result.get("errmsg", result.get("err")),
-                            result.get("code"),
-                            result)
-
-    error_msg = result.get("err", "")
-    if error_msg is None:
-        return result
-
-    if error_msg.startswith("not master"):
-        raise NotMasterError(error_msg, result)
-
-    details = result
-
-    # mongos returns the error code in an error object for some errors.
-    if "errObjects" in result:
-        for errobj in result["errObjects"]:
-            if errobj.get("err") == error_msg:
-                details = errobj
+    details = response
+    # Mongos returns the error details in a 'raw' object
+    # for some errors.
+    if "raw" in response:
+        for shard in response["raw"].values():
+            # Grab the first non-empty raw error from a shard.
+            if shard.get("errmsg") and not shard.get("ok"):
+                details = shard
                 break
 
+    errmsg = details["errmsg"]
     code = details.get("code")
+
+    # For allowable errors, only check for error messages when the code is not
+    # included.
+    if allowable_errors:
+        if code is not None:
+            if code in allowable_errors:
+                return
+        elif errmsg in allowable_errors:
+            return
+
+    # Server is "not primary" or "recovering"
+    if code is not None:
+        if code in _NOT_PRIMARY_CODES:
+            raise NotPrimaryError(errmsg, response)
+    elif HelloCompat.LEGACY_ERROR in errmsg or "node is recovering" in errmsg:
+        raise NotPrimaryError(errmsg, response)
+
+    # Other errors
+    # findAndModify with upsert can raise duplicate key error
     if code in (11000, 11001, 12582):
-        raise DuplicateKeyError(details["err"], code, result)
-    raise OperationFailure(details["err"], code, result)
+        raise DuplicateKeyError(errmsg, code, response, max_wire_version)
+    elif code == 50:
+        raise ExecutionTimeout(errmsg, code, response, max_wire_version)
+    elif code == 43:
+        raise CursorNotFound(errmsg, code, response, max_wire_version)
+
+    raise OperationFailure(errmsg, code, response, max_wire_version)
 
 
 def _raise_last_write_error(write_errors):
@@ -221,6 +188,18 @@ def _raise_write_concern_error(error):
         error.get("errmsg"), error.get("code"), error)
 
 
+def _get_wce_doc(result):
+    """Return the writeConcernError or None."""
+    wce = result.get("writeConcernError")
+    if wce:
+        # The server reports errorLabels at the top level but it's more
+        # convenient to attach it to the writeConcernError doc itself.
+        error_labels = result.get("errorLabels")
+        if error_labels:
+            wce["errorLabels"] = error_labels
+    return wce
+
+
 def _check_write_command_response(result):
     """Backward compatibility helper for write command error handling.
     """
@@ -229,20 +208,9 @@ def _check_write_command_response(result):
     if write_errors:
         _raise_last_write_error(write_errors)
 
-    error = result.get("writeConcernError")
-    if error:
-        _raise_write_concern_error(error)
-
-
-def _raise_last_error(bulk_write_result):
-    """Backward compatibility helper for insert error handling.
-    """
-    # Prefer write errors over write concern errors
-    write_errors = bulk_write_result.get("writeErrors")
-    if write_errors:
-        _raise_last_write_error(write_errors)
-
-    _raise_write_concern_error(bulk_write_result["writeConcernErrors"][-1])
+    wce = _get_wce_doc(result)
+    if wce:
+        _raise_write_concern_error(wce)
 
 
 def _fields_list_to_dict(fields, option_name):
@@ -258,10 +226,9 @@ def _fields_list_to_dict(fields, option_name):
         return fields
 
     if isinstance(fields, (abc.Sequence, abc.Set)):
-        if not all(isinstance(field, string_type) for field in fields):
+        if not all(isinstance(field, str) for field in fields):
             raise TypeError("%s must be a list of key names, each an "
-                            "instance of %s" % (option_name,
-                                                string_type.__name__))
+                            "instance of str" % (option_name,))
         return dict.fromkeys(fields, 1)
 
     raise TypeError("%s must be a mapping or "
