@@ -18,23 +18,20 @@ import datetime
 import errno
 import socket
 import struct
+import time
 
 
 from bson import _decode_all_selective
-from bson.py3compat import PY3
 
 from pymongo import helpers, message
 from pymongo.common import MAX_MESSAGE_SIZE
 from pymongo.compression_support import decompress, _NO_COMPRESSION
-from pymongo.errors import (AutoReconnect,
-                            NotPrimaryError,
+from pymongo.errors import (NotPrimaryError,
                             OperationFailure,
                             ProtocolError,
-                            NetworkTimeout,
                             _OperationCancelled)
 from pymongo.message import _UNPACK_REPLY, _OpMsg
 from pymongo.monitoring import _is_speculative_authenticate
-from pymongo.monotonic import time
 from pymongo.socket_checker import _errno_from_exception
 
 
@@ -127,7 +124,7 @@ def command(sock_info, dbname, spec, secondary_ok, is_mongos,
                 max_doc_size > max_bson_size):
             message._raise_document_too_large(name, size, max_bson_size)
     else:
-        request_id, msg, size = message.query(
+        request_id, msg, size = message._query(
             flags, ns, 0, -1, spec, None, codec_options, check_keys,
             compression_ctx)
 
@@ -192,7 +189,7 @@ def receive_message(sock_info, request_id, max_message_size=MAX_MESSAGE_SIZE):
     """Receive a raw BSON message or raise socket.error."""
     timeout = sock_info.sock.gettimeout()
     if timeout:
-        deadline = time() + timeout
+        deadline = time.monotonic() + timeout
     else:
         deadline = None
     # Ignore the response's request id.
@@ -243,7 +240,7 @@ def wait_for_read(sock_info, deadline):
                 # Wait up to 500ms for the socket to become readable and then
                 # check for cancellation.
                 if deadline:
-                    timeout = max(min(deadline - time(), _POLL_TIMEOUT), 0.001)
+                    timeout = max(min(deadline - time.monotonic(), _POLL_TIMEOUT), 0.001)
                 else:
                     timeout = _POLL_TIMEOUT
                 readable = sock_info.socket_checker.select(
@@ -252,50 +249,24 @@ def wait_for_read(sock_info, deadline):
                 raise _OperationCancelled('hello cancelled')
             if readable:
                 return
-            if deadline and time() > deadline:
+            if deadline and time.monotonic() > deadline:
                 raise socket.timeout("timed out")
 
-# memoryview was introduced in Python 2.7 but we only use it on Python 3
-# because before 2.7.4 the struct module did not support memoryview:
-# https://bugs.python.org/issue10212.
-# In Jython, using slice assignment on a memoryview results in a
-# NullPointerException.
-if not PY3:
-    def _receive_data_on_socket(sock_info, length, deadline):
-        buf = bytearray(length)
-        i = 0
-        while length:
-            try:
-                wait_for_read(sock_info, deadline)
-                chunk = sock_info.sock.recv(length)
-            except (IOError, OSError) as exc:
-                if _errno_from_exception(exc) == errno.EINTR:
-                    continue
-                raise
-            if chunk == b"":
-                raise AutoReconnect("connection closed")
+def _receive_data_on_socket(sock_info, length, deadline):
+    buf = bytearray(length)
+    mv = memoryview(buf)
+    bytes_read = 0
+    while bytes_read < length:
+        try:
+            wait_for_read(sock_info, deadline)
+            chunk_length = sock_info.sock.recv_into(mv[bytes_read:])
+        except (IOError, OSError) as exc:
+            if _errno_from_exception(exc) == errno.EINTR:
+                continue
+            raise
+        if chunk_length == 0:
+            raise OSError("connection closed")
 
-            buf[i:i + len(chunk)] = chunk
-            i += len(chunk)
-            length -= len(chunk)
+        bytes_read += chunk_length
 
-        return bytes(buf)
-else:
-    def _receive_data_on_socket(sock_info, length, deadline):
-        buf = bytearray(length)
-        mv = memoryview(buf)
-        bytes_read = 0
-        while bytes_read < length:
-            try:
-                wait_for_read(sock_info, deadline)
-                chunk_length = sock_info.sock.recv_into(mv[bytes_read:])
-            except (IOError, OSError) as exc:
-                if _errno_from_exception(exc) == errno.EINTR:
-                    continue
-                raise
-            if chunk_length == 0:
-                raise AutoReconnect("connection closed")
-
-            bytes_read += chunk_length
-
-        return mv
+    return mv
